/* Example from http://stackoverflow.com/questions/7730378/using-uncrustify-without-aligning-under-open-parenthesis */

void f(void)
{
    while (one &&
           two)
    {
        continue;
    }
}
